/*
	Copyright (C) 2004 Jonas Lindholm

	This software was developed by Jonas Lindholm, jlhm@usa.net

	History

	V1.0		Jonas Lindholm	2004-05-14

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>

#include <ldap.h>

#include <starlet.h>
#include <ssdef.h>
#include <stsdef.h>
#include <psldef.h>
#include <descrip.h>

#include <lib$routines.h>

#include <pthread.h>

#include "lgildap.h"

#include "server.h"


/*
	Description:

	This routine is invoked when the client request the supplied identity
	to be verified.
	The verification is done using the base DN defined with logical name
	LGI_LDAP_BASEDN as base. The LDAP_SCOPE_SUBTREE is used to search any children
	under the base DN for a match.
	The logical name LGI_LDAP_FILTER can be defined to filter the matches.
	Default filer is "(uid=%s)". This filter will match entries that has an uid
	that match the supplied identity (username).

	Should several entries match for the same identity the first one is used.

	If several uid's are returned the first one that match an already existing
	SYSUAF entry is used. Should there be no matching SYSUAF entry the first
	returned uid is used to create a user in SYSUAF but only if logical name
	LGI_LDAP_ADD_USER is defined.

	If the specified identity result in an uid that correspond to a SYSUAF entry
	the validity of the identity is granted. Otherwise the identity is denied.

	INPUT:		session handler
			LGI message

	OUTPUT:		If the identity is valid SS$_NORMAL
			If the identity is invalid SS$_ABORT. The message status
			will be LLSRV_NOSUCHUSER.
*/

int thread_cmd_identity(LGI_SESS_T *sess, LIM_ENTRY_T *le) {
	int		stat;
	char		basedn[256];
	char		filter[256] = "(uid=%s)";

	sess->jobtype = le->lim.request.identity.jobtype;

        le->lim.rplyid = le->lim.reqid;
        le->lim.reqid++;
	le->lim.status = LLSRV_NOSUCHUSER;

	if (!$VMS_STATUS_SUCCESS(do_trnlnm("LGI_LDAP_BASEDN", basedn, sizeof(basedn) - 1, PSL$C_EXEC))) {

		lib$stop(LLSRV_LDAPBASEDNMISS, 0);

	}

	do_trnlnm("LGI_LDAP_FILTER", filter, sizeof(filter) - 1, PSL$C_EXEC);

	pthread_mutex_lock(&ldapmtx);

	if (sess->ldap != NULL) {
		int			msgid;
		char			search[strlen(filter) + (strlen(le->lim.request.identity.identity) * 2) + 1];
		char                    *attrs[] = { "uid", NULL };
		LDAPMessage		*res;

		sprintf(search, filter, le->lim.request.identity.identity, le->lim.request.identity.identity);

	        if ((msgid = ldap_search(sess->ldap, basedn, LDAP_SCOPE_SUBTREE, search, attrs, 0)) > 0) {
			struct timeval		to = { 0, 0 };
			LDAPMessage             *e;
			LDAPMessage             *n;


			while ((stat = ldap_result(sess->ldap, msgid, 1, &to, &res)) == 0) {
				struct timespec		wait = { 0, 50000 };

				pthread_mutex_unlock(&ldapmtx);
				pthread_delay_np(&wait);
				pthread_mutex_lock(&ldapmtx);

			}

			if (stat > 0) {
				int		cnt = 0;

				for (; cnt < 2; cnt++) { /* 0 = verify without creating, 1 = create if missing */

					for ((e = ldap_first_entry(sess->ldap, res)); e != NULL;
					      n = ldap_next_entry(sess->ldap, e), ldap_msgfree(e), e = n) {

						if ((sess->dn = ldap_get_dn(sess->ldap, e)) != NULL) {
							char			**uids;

							if ((uids = ldap_get_values(sess->ldap, e, "uid")) != NULL) {
								int		idx = 0;

								while (uids[idx] != NULL) {

									pthread_mutex_unlock(&ldapmtx);

									if ($VMS_STATUS_SUCCESS(verify_username(sess, uids[idx], cnt))) {

										strncpy(le->lim.reply.identity.identity, uids[idx],
											strlen(uids[idx]));
										le->lim.reply.identity.identity[strlen(uids[idx])] = 0;

										strcpy(sess->identity, le->lim.reply.identity.identity);

										le->lim.status = SS$_NORMAL;

										sess->flags.identityok = 1;

										pthread_mutex_lock(&ldapmtx);

										ldap_value_free(uids);
										ldap_msgfree(e);
                         							ldap_msgfree(res);

										pthread_mutex_unlock(&ldapmtx);
										return SS$_NORMAL;
									}

									pthread_mutex_lock(&ldapmtx);
									idx++;
								}

								ldap_value_free(uids);
							}

							ldap_memfree(sess->dn);
							sess->dn = NULL;
						}
					}
				}

				ldap_msgfree(res);
				pthread_mutex_unlock(&ldapmtx);

       				return SS$_ABORT;
			}
		}

		ldap_unbind(sess->ldap);
		sess->ldap = NULL;
	}

	pthread_mutex_unlock(&ldapmtx);

	return SS$_ABORT;
}
