/*
	Copyright (C) 2004 Jonas Lindholm

	This software was developed by Jonas Lindholm, jlhm@usa.net

	History

	V1.0		Jonas Lindholm	2004-05-14

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
	ldap.h include all standard headers such as string.h
*/

#include "ldap.h"

#include <starlet.h>
#include <ssdef.h>
#include <stsdef.h>
#include <psldef.h>
#include <lnmdef.h>
#include <syidef.h>
#include <descrip.h>
#include <lckdef.h>

#include <lib$routines.h>
#include <libvmdef.h>

#include <rms.h>
#include <prcdef.h>
#include <pthread.h>

#include "lgi.h"
#include "lgildap.h"
#include "server.h"


pthread_mutex_t         memmtx;
pthread_mutex_t         getvmmtx;

int unsigned		qe_zone;
int			cluster = 0;
int unsigned		lockids[1];


/*
	Description:

	This routine is invoked to allocate N bytes using malloc.
	A mutex is used to controll access.
	Do not call this routine from AST.

	INPUT:		size of memory to allocate

	OUTPUT:		pointer to allocated memory
			NULL is something went wrong
*/

void *safe_malloc(int size) {
        void            *p;

        pthread_mutex_lock(&memmtx);

        p = malloc(size);

	if (p == NULL) {

		lib$stop(LLSRV_MEMALLOCFAIL, 1, size);

	}

        pthread_mutex_unlock(&memmtx);
        return p;
}


void safe_free(void *p) {
        pthread_mutex_lock(&memmtx);

        free(p);

        pthread_mutex_unlock(&memmtx);
}


void *alloc_qe(int size) {
        int             stat;
        void            *qe;

        pthread_mutex_lock(&getvmmtx);

        stat = LIB$GET_VM(&size, &qe, &qe_zone);

        pthread_mutex_unlock(&getvmmtx);

        return qe;
}


void free_qe(int size, void *qe) {
        int             stat;

        pthread_mutex_lock(&getvmmtx);

        stat = LIB$FREE_VM(&size, qe, &qe_zone);

        pthread_mutex_unlock(&getvmmtx);
}


int insert_qt(void *hdr, void *ent) {

        return LIB$INSQTIQ(ent, hdr, 0);
}


int insert_qh(void *hdr, void *ent) {

        return LIB$INSQHIQ(ent, hdr, 0);
}


void *remove_qt(void *qhd) {
        int             stat;
        void            *qe;

        stat = LIB$REMQTIQ(qhd, &qe, 0);
        return ($VMS_STATUS_SUCCESS(stat) ? qe : NULL);
}


void *remove_qh(void *qhd) {
        int             stat;
        void            *qe;

        stat = LIB$REMQHIQ(qhd, &qe, 0);
        return ($VMS_STATUS_SUCCESS(stat) ? qe : NULL);
}


void *get_free_qe(void *qhd, int size) {
        void            *qe = remove_qh(qhd);

        if (qe == NULL) {

                qe = alloc_qe(size);

        }

        return qe;
}


void init_q(void *qhd, int size, int cnt) {

	for(; cnt > 0; cnt--) {

		insert_qt(qhd, alloc_qe(size));

	}
}
                           

void create_vmzone() {
        int             stat;
        int             align = 16;
        int             flags = LIB$M_VM_GET_FILL0;

        pthread_mutex_init(&getvmmtx, NULL);

        stat = LIB$CREATE_VM_ZONE(&qe_zone, 0, 0, &flags, 0, 0, 0, &align, 0, 0, 0, 0, 0);

        if (!$VMS_STATUS_SUCCESS(stat)) { lib$stop(stat); }

}



int do_trnlnm(char *lnm, char *trn, int trnlen, int accmode) {
	int			stat;
	int unsigned		mask = LNM$M_CASE_BLIND;
	short unsigned		retlen = 0;
	ITEM3			itmlst[2] = { { trnlen, LNM$_STRING, trn, &retlen }, { 0, 0, 0, 0 } };

	$DESCRIPTOR(tbldsc, "LNM$SYSTEM_TABLE");
	$DESCRIPTOR(lnmdsc, lnm);

	lnmdsc.dsc$w_length = strlen(lnm);

	if (($VMS_STATUS_SUCCESS(stat = sys$trnlnm(&mask, &tbldsc, &lnmdsc, &accmode, &itmlst)))) {

		trn[retlen] = 0;

	}

	return stat;	
}


int translate_unix2vms(char *unixdir, char *defdev, int defdevlen, char *defdir, int defdirlen) {
        int                     cnt = 0;
        char                    *p = unixdir;
        char                    *d = defdir;
        char                    *s;

        *d++ = '[';
	
        while (*p != 0 && ((s = strchr(p, '/')) != NULL || (s = strchr(p, 0)) != NULL)) {

                if ((s - p) > 0) {
                        int		len;

                        if (cnt++ == 0) {

				len = MIN((s - p), defdevlen - 2);
                                strncpy(defdev, p, len);
                                defdev[len] = ':';
                                defdev[len + 1] = 0;

                        } else {

                                if (cnt > 2) { *d++ = '.'; };
				*d = 0;
				len = MIN((s - p), (defdirlen - strlen(defdir) - 2));
                                strncpy(d, p, len);
                                d += len;

                        }

                        p += (s - p);
                }

                p += (*p != 0);
        }

        *d++ = ']';
        *d = 0;

        return cnt;
}


LDAP *open_ldap() {
        int             stat;
        char            ldaphost[256];
        char            basedn[256];
        LDAP            *ldap = NULL;

        if (!use_ssl) {

                if (!$VMS_STATUS_SUCCESS(do_trnlnm("LGI_LDAP_HOST", ldaphost, sizeof(ldaphost) - 1, PSL$C_EXEC))) {

                        lib$stop(LLSRV_LDAPHOSTMISS);

                }
        }

        if (!$VMS_STATUS_SUCCESS(do_trnlnm("LGI_LDAP_BASEDN", basedn, sizeof(basedn) - 1, PSL$C_EXEC))) {

                lib$stop(LLSRV_LDAPBASEDNMISS);

        }

        pthread_mutex_lock(&ldapmtx);

        if ((ldap = ldap_init((use_ssl ? "127.0.0.1" : ldaphost), (use_ssl ? SSL_PORT: 389))) != NULL) {
		int             vers = 3;
		int             timeout = 10; /* allow only 10 seconds for LDAP operations */

		if ((stat = ldap_set_option(ldap, LDAP_OPT_PROTOCOL_VERSION, &vers)) == 0 &&
		    (stat = ldap_set_option(ldap, LDAP_OPT_TIMELIMIT, &timeout)) == 0) {

#if LDAP_API_VERSION > 2004
			char		dummy[256];

			if (!$VMS_STATUS_SUCCESS(do_trnlnm("LGI_LDAP_USE_NO_SSL", dummy, sizeof(dummy) - 1, PSL$C_EXEC))) {

				stat = ldap_tls_start(ldap, 1);

				if (stat != 0 &&
				    $VMS_STATUS_SUCCESS(do_trnlnm("LGI_LDAP_NO_SSL_OK", dummy, sizeof(dummy) - 1, PSL$C_EXEC))) {

					stat = 0; /* SSL failed but it is ok ! */

				}
			}
#endif

			if (stat == 0 && (stat = ldap_simple_bind_s(ldap, NULL, NULL)) == 0) {

				pthread_mutex_unlock(&ldapmtx);

                                return ldap;

			}
                }

		ldap_unbind(ldap);
        }

        pthread_mutex_unlock(&ldapmtx);

        lib$signal(LLSRV_LDAPCONNERR, 0);

        return NULL;
}


void unlock_cluster(pthread_mutex_t *mtx, int lockid) {

	if (cluster) {
		int		stat;
		int unsigned	lksb[2] = { 0, lockids[lockid] };

		stat = sys$enqw(0, LCK$K_NLMODE, &lksb, LCK$M_CONVERT, 0, 0, 0, 0, 0, 0, 0, 0);

	} else {

		pthread_mutex_unlock(mtx);

	}
}


void lock_cluster(pthread_mutex_t *mtx, int lockid) {

	if (cluster) {
		int unsigned	stat;
		int unsigned	lksb[2] = { 0, lockids[lockid] };

		stat = sys$enqw(0, LCK$K_EXMODE, &lksb, LCK$M_CONVERT, 0, 0, 0, 0, 0, 0, 0, 0);

	} else {

		pthread_mutex_lock(mtx);

	}
}




int util_init() {
	int		stat;
	ITEM3		itmlst[] = { { sizeof(cluster), SYI$_VAXCLUSTER, &cluster, 0 }, { 0, 0, 0, 0 } };

	pthread_mutex_init(&memmtx, NULL);

	create_vmzone();

	stat = sys$getsyi(0, 0, 0, &itmlst, 0, 0, 0);

	if (cluster) {
		int unsigned		lksb[2];
		$DESCRIPTOR(lcknam1dsc, "LGI_LDAP_LCK_1");

		stat = sys$enqw(0, LCK$K_NLMODE, &lksb, LCK$M_SYSTEM | LCK$M_EXPEDITE, &lcknam1dsc, 0, 0, 0, 0, 0, 0, 0);

		lockids[0] = lksb[1];
	}

	return SS$_NORMAL;
}
